#!/usr/bin/python

"""Question from reddit https://www.reddit.com/r/GAMETHEORY/comments/178eqbx/is_my_translation_into_matrix_form_correct/"""
import numpy as np
import argparse

import json
from fractions import Fraction
from game import Game

def negadd():
    payoffs = np.zeros((3,3,2))
    for a1 in [0, 1, 2]:
        for a2 in [0, 1, 2]:
            payoffs[((a1, a2, 0))] = a1 + a1 * a2 - (2 * a2 + a1 * a1)
            payoffs[((a1, a2, 1))] = a2 + a1 * a2 - (2 * a1 + a2 * a2)
    #print(payoffs.shape)
    return Game(payoffs)

if __name__ == '__main__':
    theGame = negadd()
    print(repr(theGame.find_all_equilibria()))



import numpy
from fractions import Fraction
from itertools import combinations, product

def coords_from_pos(thearray, pos):
    """Get the corrdinates from a one-dimensional position and shape. We will make the first index most signifcant."""
    # we should probably replace this with iterindices where it is used.
    coords = []
    for ii in range(len(thearray.shape) - 1):
        prod = numpy.prod(thearray.shape[1 + ii:])
        coord = pos // prod
        coords.append(coord)
        pos -= coord * prod
    coords.append(pos)
    return tuple(coords)

def mult(iterable):
    prod = 1
    for elm in iterable:
        prod *= elm
    return prod

def pos_from_coords(shape, coords):
    sum_ = 0
    for ii, coord in enumerate(coords):
        mul = mult([shape[iii] for iii in range(ii + 1, len(shape)) ])
        sum_ += coord * mul
    return sum_

def enumershape(shape):
    pieces = [[elm for elm in range(shape[ii])] for ii in range(len(shape))]
    ii = -1
    for elm in product(*pieces):
        ii += 1
        yield (ii, elm)


def iterprob(actions):
    """Generator for iterating through an array of player action probabilities. Yields a tuple (player_actions, probability) where player_actions is a tuple showing
        which player performed which action."""
    pos = [0] * len(actions)
    done = False
    while not done:
        pospos = 0
        while pos[pospos] == len(actions[pospos]) - 1:
            pospos += 1
            if pospos == len(actions):
                done = True
                break
        prob = 1
        for ii in range(len(actions)):
            prob *= actions[ii][pos[ii]]
        # print(pos, pospos)
        oldpos = tuple(pos)
        if not done:
            pos[pospos] += 1
            for ii in range(pospos):
                pos[ii] = 0
        # print('oldpos', oldpos, 'prob', prob)
        yield (oldpos, prob)
        
def itersupport(support_actions):
    """Generator for iterating through support_actions. Support_actions is a list of lists, first index in player. Inner list is a tuple, (player_action, probility).
       probability is either a symbol or a number
       yields a tuple (player_actions, probability). Probability is either a number or a symbol expression."""
    pos = [0] * len(support_actions)
    done = False
    while not done:
        pospos = 0
        while pos[pospos] == len(support_actions[pospos]) - 1:
            pospos += 1
            if pospos == len(support_actions):
                done = True
                break
        prob = 1
        palist = []
        for ii in range(len(support_actions)):
            # print(ii, support_actions[ii][pos[ii]])
            palist.append( support_actions[ii][pos[ii]][0])
            prob *= support_actions[ii][pos[ii]][1]
        # print(pos, pospos)
        if not done:
            pos[pospos] += 1
            for ii in range(pospos):
                pos[ii] = 0
        # print('oldpos', oldpos, 'prob', prob)
        yield (tuple(palist), prob)
    

def iterindices(shape):
    """Iterate over possible indices given a shape tuple."""
    pos = [0] * len(shape)
    done = False
    while not done:
        pospos = 0
        while pos[pospos] == shape[pospos] - 1:
            pospos += 1
            if pospos == len(shape):
                done = True
                break
        oldpos = tuple(pos)
        if not done:
            pos[pospos] += 1
            for ii in range(pospos):
                pos[ii] = 0
        yield oldpos
        
def subsets(aset):
   """Yield all subsets of an iterable (except the empty set)"""
   for alen in range(1, len(aset) + 1):
       for acombo in combinations(aset, alen):
           yield(acombo)
        
def iter_subset_combos(actions):
    """Generator to find combinations of subsets of a list of lists.
       Yields a list of lists."""
    pos = [0] * len(actions)
    action_subsets = []
    for i in range(len(actions)):
        #print(i, actions[i], list(subsets(actions[i])))
        action_subsets.append(list(subsets(actions[i]))) 
    done = False
    while not done:
        pospos = 0
        while pos[pospos] == len(action_subsets[pospos]) - 1:
            pospos += 1
            if pospos == len(actions):
                done = True
                break
        prob = 1
        sslist = []
        for ii in range(len(actions)):
            # print(ii, support_actions[ii][pos[ii]])
            sslist.append( action_subsets[ii][pos[ii]])
        # print(pos, pospos)
        if not done:
            pos[pospos] += 1
            for ii in range(pospos):
                pos[ii] = 0
        # print('oldpos', oldpos, 'prob', prob)
        yield sslist
        
def is_pure(profile):
    """Is the profile (list of lists) pure (each player is playing exactly one stratgy)? Returns a boolean."""
    for elm in profile:
        if len(elm) != 1:
            return False
    return True
    
def dict_to_list(adict):
    """Turn a dict into a list of pairs"""
    return [[key, adict[key]] for key in adict]

def list_to_dict(alist):
   """Turn a list of list pairs into a dict"""
   return {elm[0]:elm[1] for elm in alist}
    
    

def silly(pos):
    """A silly test function that gives a result based on pos. Pos is a list of ints."""
    #print('silly')
    result = 0
    for ii in range(len(pos)):
        result += (10 ** ii) * (pos[ii] + 1)
    return result


def test_pos_from_coords():
    shape = (3,3,2)
    for ii in range(3):
        for iii in range(3):
            for iv in range(2):
                print(pos_from_coords(shape, (ii, iii, iv)))



"""A class for a multi-player normal form game"""
import traceback
from copy import deepcopy
import numpy as np
from sympy import symbols
from sympy.core import expr
from sympy import preorder_traversal
from sympy.solvers import solve
#from sympy.core.expr import Expr
from sympy.core import Number
from sympy.core import Symbol
from sympy.core import Expr
from sympy.core.numbers import Float as FloatType

from pymnash.util import iterprob, iterindices, itersupport, iter_subset_combos, is_pure, dict_to_list, list_to_dict

class Game(object):
    """ A class for a multi-player normal form game."""


    def __init__(self, payoffs, player_labels = None, action_labels = None):
        """Payoffs is an np.array of floats, giving payouts to all players.
           If labels are omitted or incomplete we'll just fill them in with stringified ints."""
        # For now I'm not using player/acion labels.
        self.verbose = False
        if type(payoffs) != np.ndarray:
            raise Exception('Payoffs must be numpy array')
        self.payoffs = payoffs
        self.player_count = self.num_players()
        player_count = self.player_count
        if player_labels is None:
            self.player_labels =  [str(ii) for ii in range(player_count)]
        elif len(player_labels) >= player_count:
            self.player_labels = list(player_labels[0:player_count])
        else:
            self.player_labels = list(player_labels)
            for ii in range(player_count - len(player_labels)):
                self.player_labels.append(str(ii + len(player_labels)))
        action_count = payoffs.shape[0] # TODO: support different action lables for different players
        if action_labels is None:
             self.action_labels = [str(ii) for ii in range(action_count)]
        elif len(action_labels) >= action_count:
            self.action_labels = list(action_labels[0:action_count])
        else:
            self.action_labels = list(action_labels)
            for ii in range(action_count - len(action_labels)):
                self.action_labels.append(str(ii + len(action_labels)))
        self._wiggle = 0.000001
        self.dominated = [[] for ii in range(self.num_players())]

    def __repr__(self):
         return 'payoffs {}\nplayer_labels {}\naction_labels {}'.format(self.payoffs,
                self.player_labels, self.action_labels)

    def eq(self, val1, val2):
        """Check whether val1 and val2 are 'close enough' to count as equal."""
        return val1 + self._wiggle > val2 and val2 + self._wiggle > val1

    def gt(self, val1, val2):
        """Check whether val1 > val2, giving ourselves a little 'wiggle room' for rounding errors."""
        return val1 > val2 + self._wiggle

    def get_profile_payoffs(self, profile):
        """Find all player payoffs given the strategy profile. Returns a list of floats."""
        payoffs = [0] * self.player_count
        for acombo in itersupport(profile):
            combo_actions = acombo[0]
            for player in range(self.player_count):
                payoffs[player] += self.payoffs[(combo_actions)][player] * acombo[1]
        return payoffs

    def is_dominated(self, profile, profile_payoffs=None):
        """Check if there exists a pure strategy for at least one player which gives that player a payoff
           than the specified profile."""
        if profile_payoffs is None:
            profile_payoffs = self.get_profile_payoffs(profile)
        for player in range(self.player_count):
            old_player_profile = profile[player]
            for anaction in range(self.payoffs.shape[player]):
                action_payoff = 0
                profile[player] = [[anaction, 1]]
                for acombo in itersupport(profile):
                    combo_actions = acombo[0]
                    action_payoff +=  self.payoffs[(combo_actions)][player] * acombo[1]
                profile[player] = old_player_profile
                if action_payoff > profile_payoffs[player] + self._wiggle:
                    if self.verbose:
                        print("is_dominated profile {} is dominated for player {} by pure strategy {}".format(profile,
                          player, anaction))
                    return True
        return False

    def is_nash(self, profile):
        """Check if the supplied strategy profile is a nash equilibrium.
           Profile is a list of lists, each list is strategy profile for one player.
           Returns a boolean."""
        # In order to be a valid nash equilibrium, each player must be indifferent
        # as to which action in his support he plays
        # (suppport is actions played with non-zero probability)
        # also, a player must not be able to do better by playing an action not in his suport.
        # of course, no probability can be negative and all probabilities must sum to 1.
        is_nash = True
        shape = self.payoffs.shape
        for player, player_profile in enumerate(profile):
            prob_sum = 0
            support_utility = None
            nonsupport_utility = None
            for prob in player_profile:
                if prob < 0:
                    raise Exception('negative probability')
                prob_sum += prob
            if not self.eq(prob_sum, 1):
                if self.verbose:
                    print('prob_sum', prob_sum)
                raise Exception('probabilities do not sum to 1')
            others_profile = deepcopy(profile)
            del others_profile[player]
            # check the utility
            for jj, prob in enumerate(player_profile):
                if self.verbose:
                    print('player', player, 'others_profile', others_profile)
                in_support = prob > 0
                myslice = [slice(None)] * len(shape)
                myslice[-1] = player
                relevant_payoffs = self.payoffs[tuple(myslice)]
                if self.verbose:
                    print("all payoffs for player", player, relevant_payoffs)
                myslice = [slice(None)] * (len(shape) - 1)
                myslice[player] = jj
                relevant_payoffs = relevant_payoffs[tuple(myslice)]
                if self.verbose:
                    print("payoffs for player", player, relevant_payoffs)
                utility = 0
                for (thetuple, theprob) in iterprob(others_profile):
                    utility += relevant_payoffs[thetuple] * theprob
                if self.verbose:
                    print('player', player, 'action', jj, 'utility', utility, 'in_support', in_support)
                is_nash = True
                if in_support:
                    if nonsupport_utility is not None and self.gt(nonsupport_utility, utility):
                        is_nash = False
                    if support_utility is None:
                        support_utility = utility
                    elif not self.eq(support_utility, utility):
                         is_nash = False
                else:
                    if support_utility is not None and self.gt(utility, support_utility):
                        is_nash = False
                    if nonsupport_utility is None or utility > nonsupport_utility:
                        nonsupport_utility = utility
                if not is_nash:
                     if self.verbose:
                        print('rejected utility', utility, 'support_utility', support_utility,
                              'nonsupport_utility', nonsupport_utility)
                     return False

        return is_nash

    def num_actions(self, player):
        """Return the number of available actions for the player with given index. Returns an int."""
        return self.payoffs.shape[player]

    def num_players(self):
        """Return the number of players for this game"""
        return len(self.payoffs.shape) - 1


    def find_pure(self):
        """Find any pure nash equilibria for this game. Returns a list of lists, one entry per equilibrium found.
           Inner list is the actions for each player."""
        # Iterate though the list of pure stretegies, check if it is a nash equilibrium
        eq = []
        shape = self.payoffs.shape[:-1]
        for indices in iterindices(shape):
             # convert indices into an action profile playing the pure actions at that index
             profile = [0] * len(shape)
             for player, action in enumerate(indices):
                 profile[player] = [0] * self.num_actions(player)
                 profile[player][action] = 1
             is_nash = self.is_nash(profile)
             print('profile', profile, 'is_nash', is_nash)
             if is_nash:
                 eq.append(indices)
        return eq


    def iesds(self):
        """Perform iteratated elimination of strictly dominated strategies to get a reduced game.
           Returns None, updates self.dominated in place"""
        progress = True
        while progress:
            progress = False
            if self.iesds1():
                progress = True
            if self.iesds2():
                progress = True


    def iesds1(self):
        """Perform iteratated elimination of strictly dominated strategies to get a reduced game,
           considering stratgies dominated by a single other strategy.
           Updates dominated in place.
           returns a boolean indicating it found at least 1 new dominated strategy. """
        real_progress = False
        progress = True
        dominated = self.dominated
        num_players = self.num_players()
        while progress:
            progress = False
            for player in range(num_players):
                for action0 in range(self.payoffs.shape[player]):
                    if action0 in dominated[player]:
                         continue
                    aslice = [slice(None)] * len(self.payoffs.shape)
                    aslice[-1] = player
                    aslice[player] = action0
                    action0_payoffs = self.payoffs[tuple(aslice)]
                    for action1 in range(self.payoffs.shape[player]):
                        skip = False
                        if action1 in dominated[player]:
                            skip = True
                            continue
                        if action0 == action1:
                            skip = True
                            continue
                        aslice = [slice(None)] * len(self.payoffs.shape)
                        aslice[-1] = player
                        aslice[player] = action1
                        action1_payoffs = self.payoffs[tuple(aslice)]
                        adominated = True
                        for indices in iterindices(action0_payoffs.shape):
                            skip2 = False
                            for ii, index in enumerate(indices):
                                inner_player = ii
                                if inner_player >= player:
                                    inner_player += 1
                                if index in dominated[inner_player]:
                                    skip2 = True
                                    break
                            if skip2:
                                 continue
                            ut0 = action0_payoffs[indices]
                            ut1 = action1_payoffs[indices]
                            if not self.gt(ut0, ut1):
                                adominated = False
                                if self.verbose:
                                    pass
                                    # print('player', player, 'action0', action0, 'action1', action1, 'u0', ut0, 'u1', ut1)
                                break
                        if (not skip) and adominated:
                             dominated[player].append(action1)
                             progress = True
                             real_progess = True
                             if self.verbose:
                                  print('player', player, 'action', action1, 'dominated by', action0)
        return real_progress

    def iesds2(self):
        """Perform iteratated elimination of strictly dominated strategies to get a reduced game,
           considering stratgies dominated by a linear combo of 2 other strategies.
           Returns a boolean indicating it found at least one new dominated strategy"""
        # For strategy 0 to be dominated by a combo of strategies 1 and 2, for every combo of other players'
        # strategies it must be the case that at least one of
        # strategies 1 and 2 score better than strategy 0 at every point. If both perform better, we don't get any
        # new information as to what combinations perform better, but if
        # only strategy 1 performs better we have a minimum ratio of strategy 1 to strategy
        # 2 for a dominating combo, and if only stratgey 2 perfoms better we get a maximum ratio.

        num_players = self.num_players()
        progress = True
        real_progress = False
        dominated = self.dominated
        while progress:
            progress = False
            for player in range(num_players):
                for action_a in range(self.payoffs.shape[player]):

                    if action_a in dominated[player]:
                        continue
                    for action_b in range(self.payoffs.shape[player]):
                        if action_b in dominated[player]:
                             continue
                        if action_b == action_a:
                             continue
                        for action_c in range(action_b + 1, self.payoffs.shape[player]):
                            if action_c in dominated[player]:
                                continue
                            if action_c == action_a:
                                continue
                            if self._combo_dominates(player, action_a, action_b, action_c):
                                progress = True
                                real_progress = True
                                dominated[player].append(action_a)
        return real_progress

    def _combo_dominates(self, player, strat_a, strat_b, strat_c):
        """Helper function for iesds2. Checks if there is some combo of strategies b and c that dominates strategy a.
           returns a boolean"""
        dominated = self.dominated
        min_p = None # minimum probability of playing b which might dominate a
        max_p = None
        pslice = [slice(None)] * len(self.payoffs.shape) # just interested in payoffs for player
        pslice[-1] = player
        sslice = deepcopy(pslice)
        sslice[player] = strat_a
        a_payoffs = self.payoffs[tuple(sslice)]
        sslice = deepcopy(pslice)
        sslice[player] = strat_b
        b_payoffs = self.payoffs[tuple(sslice)]
        sslice = deepcopy(pslice)
        sslice[player] = strat_c
        c_payoffs = self.payoffs[tuple(sslice)]
        for indices in iterindices(a_payoffs.shape):
            skip2 = False
            for ii, index in enumerate(indices):
                inner_player = ii
                if inner_player >= player:
                    inner_player += 1
                    if index in dominated[inner_player]:
                        skip2 = True
                        break
                    if skip2:
                        continue
                ut_a = a_payoffs[indices]
                ut_b = b_payoffs[indices]
                ut_c = c_payoffs[indices]
                if self.gt(ut_a, ut_b) and self.gt(ut_a, ut_c):
                    return False
                if self.gt(ut_b, ut_a) and self.gt(ut_c, ut_a):
                    continue
                if self.eq(ut_b, ut_c):
                    continue
                p = (ut_a - ut_c) / (ut_b - ut_c) # probability b at which the combo scores the same as playing a
                if ut_b > ut_c:
                    if min_p is None or p > min_p:
                        min_p = p
                else:
                    if max_p is None or p < max_p:
                        max_p = p
                    if min_p is not None and max_p is not None and min_p > max_p:
                        return False
        if self.verbose:
            print("player", player, "strategy", strat_a, "dominated by ", strat_b, strat_c, min_p, max_p)
        return True

    def _get_indifference_probs(self, support):
        """Find the combinations of probabilities such that each player is indifferent to which action in his
           own support which he plays given the probabilities of the other players.
           Support is a list of lists, players and actions. Each player could have any number of actions,
           the number of possible actions  will vary by player.
           Returns a list of dicts."""
        support_symbols = [] # list of lists. Value is a tuple (player_action (int), symbol)
        symbols_list = [] # put all symbols in one list for solver
        for player in range(len(support)):
            support_symbols.append([])
            for action in support[player]:
                name = 'prob_{}_{}'.format(player, action)
                asymbol = symbols(name)
                support_symbols[player].append((action, asymbol))
                symbols_list.append(asymbol)
        symbol_type = type(symbols_list[0])

        # print(repr(support_symbols))
        # probabilities for each player action must equal one
        psums = [] # list of probability sum equations
        for player in range(len(support)):
            eq = -1
            for action in support_symbols[player]:
                eq += action[1]
            psums.append(eq)
        #print(repr(psums))
        indiff_equations = []
        for player in  range(len(support_symbols)):
            if len(support_symbols[player]) == 1: # player is sticking with 1 strategy, no indifference for him
                continue
            ppayoffs = [] # each element is the expected payoff to the chosen player for the chosen action
            old_ss = support_symbols[player]
            for player_action in support[player]:
                support_symbols[player] = [(player_action, 1)]
                apayoff = 0
                for acombo in itersupport(support_symbols):
                    combo_actions = acombo[0]
                    apayoff += self.payoffs[(combo_actions)][player] * acombo[1]
                ppayoffs.append(apayoff)
            for payoff_index in range(1, len(ppayoffs)):
                an_equality = ppayoffs[0] - ppayoffs[payoff_index]
                indiff_equations.append(an_equality)
            support_symbols[player] = old_ss
        all_equations = psums + indiff_equations
        try:
             initial_solutions = solve(all_equations, symbols_list)
        except Exception:
             # This means there are no solutions with the given support
             #print("sympy threw an exception, here the equations ans symbols", all_equations, symbols_list)
             return []
        if not initial_solutions:
            return []
        if type(initial_solutions) == dict:
            support_result = self._sympy_dict_to_profile(initial_solutions)
            for aprofile in support_result:
                for val in aprofile.values():
                    #print('type', type(val))
                    if type(val) == FloatType:
                        if float(val) > 1 or float(val) <= 0:
                            return []
            #print(" a dict?? WFTT???")
            #print([initial_solutions])
            #print(support_result)
            return  [support_result]
        real_solutions = []
        for asolution in initial_solutions:
            support_result = [ {} for ii in range(self.player_count)]
            ok = True
            for val in asolution:
                try:
                    if type(val) != symbol_type and val > 1 or val <= 0:
                        ok = False
                        break

                except Exception:
                    pass # can't compare symbol with number, carry on
            if ok:
                sd = {}
                for ii, elm in enumerate(symbols_list):
                    name = str(elm)
                    pieces = name.split('_')
                    player = int(pieces[1])
                    action = int(pieces[2])
                    support_result[player][action] = asolution[ii]
                real_solutions.append(support_result)
        #for asolution in real_solutions:
        #    print(asolution)
        return real_solutions

    def _sympy_dict_to_profile(self, pd):
        """Given a sympy dictionary indicating solutions to indifference equations, return a support structure."""
        #The probabilities in the profile here might be a number or an expression
        profile = [ {} for ii in range(self.player_count)]

        for elm in pd.keys():
            name = str(elm)
            pieces = name.split('_')
            player = int(pieces[1])
            action = int(pieces[2])
            profile[player][action] = pd[elm]
        # The profile also must include self-referntial values for symbolic probabilities e.g
        # if sympy says player 0 has an action profile of 0: 1 - p_0_1 the we must also include in his
        # action dict a value 1: p_0_1
        for elm in pd.values():
            if isinstance(elm, Expr):
                for arg in  preorder_traversal(elm):
                    if isinstance(arg, Symbol):
                        name = str(arg)
                        pieces = name.split('_')
                        player = int(pieces[1])
                        action = int(pieces[2])
                        profile[player][action] = arg
        return profile

    def _carnate_profile(self, profile):
        """Given a support dict which may contain sympy expressions as values, return a suport with float values."""
        # We need this to evaluate the payoff to a player of a strategy profile that is an expression rather
        # than a number e.g. player 0 will play actions 0 and 1 in any probabilities that add to 1.
        # We will treat such a case as the plyer choosing all actions in the support with equal probability.
        profile_result = [{} for ii in range(self.player_count)]
        for player_index, player_actions in enumerate(profile):
            even = 1.0 / len(player_actions)
            for action in player_actions:
                action_prob = player_actions[action]
                if isinstance(action_prob, Number):
                    profile_result[player_index][action] = float(action_prob)
                elif isinstance(action_prob, Expr):
                    profile_result[player_index] = {anaction: even for anaction in player_actions}
                    continue
                else:
                    profile_result[player_index][action] = float(action_prob)
        return profile_result


    def find_all_equilibria(self):
        """Attempt to find all nash equilibria for a game. Returns a list of dicts, keys are symbols,
           values are probabilities (numbers or symbols)."""
        result = []
        action_shape = self.payoffs.shape[:-1]
        possible_actions = [list(range(player_actions)) for player_actions in action_shape]
        #print(possible_actions)
        # we should first eliminate dominated strategies. we will skip that step for now
        for acombo in iter_subset_combos(possible_actions):
           if is_pure(acombo):
               # print(acombo)
               profile = [[[player_action[0], 1]] for player_action in acombo]
               if not self.is_dominated(profile):
                    profile_dict = [list_to_dict(player_profile) for player_profile in profile]
                    result.append(profile_dict)
           else:
               #print("checking combo", acombo)
               combo_solutions = self._get_indifference_probs(acombo)
               for asol in combo_solutions:
                   # print('asol', asol)
                   #listy = [dict_to_list(psol) for psol in asol]
                   #print('psol', psol)
                   carnate = self._carnate_profile(asol)
                   listy = [dict_to_list(elm) for elm in carnate]
                   if not self.is_dominated(listy):
                       result.append(asol)
        return result



#!/usr/bin/python

"""Functions to create some sample test games"""

import numpy as np
import argparse

import json
from pymnash.game import Game
from pymnash.util import coords_from_pos, iterindices, enumershape
from collections import defaultdict


def battle_of_genders(n):
    """A generalization of the 'battle of the sexes' game which supports any number of genders.
    The number of possible moves for each player iss equal to the total number of players.
    All players get zero payoff unless they all coordinate on a choice, each player has a particular favorite choice
    but all players get something as long as they all coordinate."""
    # We know there is one pure strategy per player, with all players playing that player's favorite.
    # There will also be one mixed  strategy for ever combination of pure strategies
    boga = np.zeros(tuple([n] * (n + 1)), dtype=float)
    #print(boga)
    for ii in range(n):
        loc = boga
        for jj in range(n):
            loc = loc[ii]
        for kk in range(n):
            if kk == ii:
                loc[kk] = n
            else:
                 loc[kk] = 1
    return Game(boga)

def reducible(n):
    """A game just to test the iterated elimination of strictly dominated strategies. """
    # 2 player version scored come from steven tadelis game theory an introduction.
    # I will extend it with obvious dominant strategies of other players.
    twoplayer = [[[4,3], [5,1], [6,2]],
                 [[2,1], [8,4], [3,6]],
                 [[3,0], [9,6], [2,8]]]
    thetuple = tuple([3, 3] + [2] * (n - 2) +  [n])
    payoffs = np.zeros(thetuple)
    for indices in iterindices(payoffs.shape):
        player = indices[len(indices) - 1]
        if player in [0, 1]:
             payoffs[indices] = twoplayer[indices[0]][indices[1]][player]
        elif indices[player] == 0:
             payoffs[indices] = 1
    return Game(payoffs)

def combo_reducible(n):
    """A game for testing iesds2. It has a strategy that is not dominated by any single other strategy
       but can be dominated by a linear combination of strategies."""
    # The game comes from steven tadelis game theory an introduction, page 115 of first edition.
    if n != 2:
        raise Exception('This game is only supported for two players')
    payoffs = [[[5, 1], [1,4], [1, 0]],
               [[3,2], [0, 0], [3, 5]],
               [[4,3], [4,4], [0,3]]]
    payoffs = np.array(payoffs)
    return Game(payoffs)



def dunderheads(n):
    """Multi-player High-Low. Similar to battle of genders, except there are only two options and all players
        have the same preferrred option."""
    # There are two pure nash equilibria, the 'smart' on where evryone picks the preferred option, and the
    # 'dunderheaded' one where everyone picks the option they don't like.
    # There should also be a 'super-dunderheaded' solutions where all player mix their picks.
    # The probability of pcking the dunderheaded option will be higher, so the higher chance of getting a match
    # exactly compensates for the lower payoff.
    payoffs = np.zeros(tuple([2] * n + [n]), dtype=float)
    for indices in iterindices(payoffs.shape):
        good = True # everyone is playing the good choice
        bad = True
        for ii in range(len(indices) - 1):
            if indices[ii] == 1: #someone is playing the bad choice
                good = False
            elif indices[ii] == 0:
                bad = False
        if good:
            payoffs[indices] = 3.0
        if bad:
            payoffs[indices] = 1.0
    return Game(payoffs)



def prisoners_dilemma(n):
    """For the multiplayer prisoner's dilemma we will say if everone cooperates then each person gets a -1 payoff.
       If exactly one player defectes he gets 0 and everyone else gets -5.
       If multiple player defect, the ones that defect get -3 and the ones that cooperate get -5.
    """
    # We will say for ever player choice '0' is cooperate and '1' is defect.
    # The innermost array is player id for payoffs
    # We know the only nash equilibrium is everybody defects.
    payoffs = np.zeros(tuple(([2] * n + [n])), dtype=float)
    for ii in range(np.prod(payoffs.shape)):
        coords = coords_from_pos(payoffs, ii)
        player = coords[ n ] # index of the player we are looking at
        he_defected = bool(coords[player])
        total_defected = sum(coords[:-1])
        if total_defected == 0:
            payoffs[coords] = -1
        elif he_defected and total_defected  == 1:
            payoffs[coords] = 0
        elif he_defected:
             payoffs[coords] = -3
        else:
             payoffs[coords] = -5
    return Game(payoffs, prisoner_labels)


def matching_pennies(n):
    """For n player matching pennies each player can choose a number in the range 0 to n -1, and player m
       wins if the result is m modulo n.
       For the 2 x 2 case think of heads as zero and tails as 1, so player 0 wins with HH or TT,
       player 1 wins with HT or TH"""
    # The obvious equilibrium is everyone plays randomly. But there are many possibilities for multiple players.
    # In the 3 player version, if 2 players
    # play randomly, the other player can play anything and it's still an equilibrium.
    # As long as at least one player is playing randomly it doesn't really matter what the
    # other players do, but just one player randomising would not be an equilibrium.
    payoffs = np.zeros(tuple([n] * (n + 1)), dtype=float)
    for ii in range(np.prod(payoffs.shape)):
        coords = coords_from_pos(payoffs, ii)
        player = coords[ n ] # index of the player we are looking at
        sum_played = sum(coords[:-1])
        if sum_played % n == player:
            payoffs[coords] = n - 1
        else:
            payoffs[coords] = -1
    #print(payoffs)
    return Game(payoffs)

def how_low_dare_you_go(n, m):
    """n players have m choices of numbers, m > n. The winning player is the player who picks
       the lowest non-negative integer not chosen by any other player."""
    # we will restrict the game to m choices so we have a hope of finding solutions.
    # In principle this game could be played with an infinite number of possible
    # moves, nobody igoing to play a number all that much higher than the number of players in any case.
    payoffs = np.zeros(tuple(([m] * n + [n])), dtype=float)
    for jj, coords in enumershape(payoffs.shape[:-1]):
        counts = defaultdict(int) # number of players picking this number
        win = None
        for iii in range(len(coords)):
            counts[coords[iii]] += 1
        for ii in sorted(counts.keys()):
            if counts[ii] == 1:
                win = ii
                break
        if win is not None:
            done = False
            for player_index in range(len(coords)):
                if coords[player_index] == win:
                    pos = list(coords)
                    pos.append(player_index)
                    payoffs[tuple(pos)] = 1
    return Game(payoffs)

def mixed_dom(n, m):
    """A two-player game where one player has a dominated strategy, but it takes a combination of
       m strategies to defeat it (one player has m strategies, the other has m + 1).
       Based on a comment by kevinwangg in the thread
       https://www.reddit.com/r/GAMETHEORY/comments/18d5zxx/dominated_by_3_or_more_strategies/
    """
    if n != 2:
        raise ValueError("This game is only supported for 2 players")
    payoffs = np.zeros(tuple((m + 1, m, 2)), dtype=float)
    for ii in range(m):
        payoffs[ii][ii][0] = m + 1
        payoffs[ii][ii][1] = -1 * (m + 1)
        payoffs[m][ii][0] = 1
        payoffs[m][ii][1] = -1
    return Game(payoffs)

def chicken(n):
    """For this variation of multi-player cheicken, each player has 2 possible actions
       (0 = chicken, 1 = hawk).  If everyne is a chicken, everyone scores 0.
       If there is just one hawk, he gets 5 points and the chickens each lose 1.
       If there is more than 1 hawk, they each score -10 and the chickens get zero.
    """
    payoffs = np.zeros(tuple([2] * n + [n]), dtype=float)
    shape = payoffs.shape[:-1]
    for pos in iterindices(shape):
        hawks = 0
        for ii in range(len(shape)):
            if pos[ii] == 1:
                hawks += 1
        for ii in range(n):
            where = tuple(list(pos) + [ii])
            if hawks == 1:
                if pos[ii] == 1:
                    payoffs[where] = 5
                else:
                    payoffs[where] = -1
            elif hawks > 1:
                if pos[ii] == 1:
                    payoffs[where] = -10
    return Game(payoffs)

def stag_hunt(n, m):
    """Each player has 2 options, 0 = hunt the stag, 1 = chase rabbits.
       We need a critical number m of stag hunters to catch the stag.
       If an insufficient number goes after the stag, they get 0
       If a sufficient number go after the stag, they share the value which I will say is 4 * n.
    """
    payoffs = np.zeros(tuple([2] * n + [n]), dtype=float)
    shape = payoffs.shape[:-1]
    for pos in iterindices(shape):
        hunters = 0
        for ii in range(len(shape)):
            if pos[ii] == 1:
                hunters += 1
        for ii in range(n):
            where = tuple(list(pos) + [ii])
            if pos[ii] == 0:
                payoffs[where] = 1
            else:
                if hunters >= m:
                    payoffs[where] = 4 * n / hunters
    return Game(payoffs)


game_names = {"battle_of_genders":battle_of_genders, "reducible":reducible, "combo_reducible":combo_reducible,
               "dunderheads":dunderheads, "prisoners_dilemma":prisoners_dilemma, "matching_pennies":matching_pennies,
              "how_low_dare_you_go":how_low_dare_you_go, "mixed_dom":mixed_dom, 'chicken':chicken,
              'stag_hunt':stag_hunt}


def get_game_fun(name):
    """Find the factory function based on the game name"""
    if name in game_names:
        return game_names[name]
    for game_name in game_names.keys():
        if game_name.startswith(name):
            return game_names[game_name]
    raise ValueError("Unknown game {}".format(name))

def get_canned_profile(fun, num_players):
    """Return a sample strategy profile for the game."""
    if fun == battle_of_genders:
        inner = [0] * num_players
        profile = [list(inner) for ii in range(num_players)]
        for ii in range(num_players):
            profile[ii][ii] = 1
        return profile
    if fun == dunderheads:
        aroot = 3.0 ** (1/(num_players - 1.0))
        profile = [[1 / (aroot + 1), aroot/(aroot + 1)]] * num_players
        return profile
    if fun == reducible:
        profile = [[1, 0, 0]] * 2 +  [[1, 0]] * (num_players - 2)
        return profile
    if fun == matching_pennies:
        frac = 1 / num_players
        profile = [[frac] * num_players,
                   [frac] * num_players] + [[1] + [0] * (num_players - 1)] * (num_players - 2)
        return profile

    raise Exception("Unsupported")

if __name__ == '__main__':
    from argparse import ArgumentParser
    from ast import literal_eval
    parser = ArgumentParser()
    parser.add_argument('--players', help = "number of players", type=int, default=3)
    parser.add_argument('--m', type=int, default=None, help="integer parameter that exists for some games")
    parser.add_argument('--game', help="game name", default='battle_of_genders')
    parser.add_argument('--pure', action='store_true', help='find pure strategy equilibria')
    parser.add_argument('--payoffs', help='show payoff matrix', action='store_true')
    parser.add_argument('--profile', help='test if profile is equilibrium')
    parser.add_argument('--canned', help='test if canned profile is equilibrium', action='store_true')
    parser.add_argument('--verbose', action='store_true')
    parser.add_argument('--iesds', action='store_true')
    parser.add_argument('--combo', action='store_true', help='check if a combo of strategies dominates a strategy')
    parser.add_argument('--support', help='try to find nash equilibria with the given support', default = None)
    parser.add_argument('--all', help='try to find all nash equilibria for this game', action='store_true')
    args = parser.parse_args()
    agame = None
    profile = None
    game_fun = get_game_fun(args.game)
    if game_fun in [how_low_dare_you_go, mixed_dom, stag_hunt]:
        agame = game_fun(args.players, args.m)
    else:
        agame = game_fun(args.players)

    if args.payoffs:
        print('payoffs:')
        print(repr(agame.payoffs))
        print('')
    if args.pure:
        print('pure:')
        print(repr(agame.find_pure()))
    if args.canned:
        profile = get_canned_profile(game_fun, args.players)
        print('profile:')
        print(profile)
        is_nash = agame.is_nash(profile)
        print('is_nash:', is_nash)
    if args.profile:
        profile = literal_eval(args.profile)
        print('profile:')
        print(profile)
        is_nash = agame.is_nash(profile)
        print('is_nash:', is_nash)
    if args.iesds:
        agame.iesds()
        print('dominated strategies:')
        print(agame.dominated)
    if args.combo:
        thecombo = literal_eval(args.combo)
        print(agame._combo_dominates(1, 0, 1, 2))
    if args.support:
        support = literal_eval(args.support)
        indifference_probs = agame._get_indifference_probs(support)
        print('support:')
        print(support)
        print('indifference probs:')
        print(indifference_probs)
    if args.all:
        all_nash = agame.find_all_equilibria()
        for anash in all_nash:
            print(anash)

#./test_sample_game.py --game battle --support "[[0,1,2], [0,1,2], [0,1,2]]" # this will give unique probs
#./test_sample_game.py --game battle --support "[[1,2], [1,2], [1,2]]"
#./test_sample_game.py --game battle --players 4 --support "[[1,2], [1,2], [0,3], [0, 3]]" # this should give lines of solutions.
